{
 "cells": [
  {
   "cell_type": "markdown",
   "source": [
    "Learning Embeddings with Continuous Bag of Words (CBOW)\r\n",
    "=========================================================\r\n",
    "\r\n",
    "The pipeline from text to vectorized minibatch is mostly standard: both the Vocabulary and the DataLoader function exactly as they did in [Example: Classifying Sentiment of Restaurant Reviews](https://weiliu2k.github.io/CITS4012/classification/yelp.html). However, unlike the Vectorizers we saw there, the Vectorizer in this section does not construct onehot vectors. Instead, a vector of integers representing the indices of the context is constructed and returned."
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "## Imports"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "source": [
    "import os\r\n",
    "from argparse import Namespace\r\n",
    "from collections import Counter\r\n",
    "import json\r\n",
    "import re\r\n",
    "import string\r\n",
    "\r\n",
    "import numpy as np\r\n",
    "import pandas as pd\r\n",
    "import torch \r\n",
    "import torch.nn as nn\r\n",
    "import torch.nn.functional as F\r\n",
    "import torch.optim as optim\r\n",
    "from torch.utils.data import Dataset, DataLoader\r\n",
    "from tqdm import tqdm_notebook\r\n",
    "from tqdm.notebook import tqdm"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "## Data Vectorization classes (from text to vectorized minibatches)"
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### The Vocabulary\r\n",
    "\r\n",
    "A `Vocabulary` object allows the lookup of index by a token, and the lookup of token by index, and add tokens to the vocabulary. "
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "source": [
    "class Vocabulary(object):\r\n",
    "    \"\"\"Class to process text and extract vocabulary for mapping\"\"\"\r\n",
    "\r\n",
    "    def __init__(self, token_to_idx=None, mask_token=\"<MASK>\", add_unk=True, unk_token=\"<UNK>\"):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            token_to_idx (dict): a pre-existing map of tokens to indices\r\n",
    "            mask_token (str): the MASK token to add into the Vocabulary; indicates\r\n",
    "                a position that will not be used in updating the model's parameters\r\n",
    "            add_unk (bool): a flag that indicates whether to add the UNK token\r\n",
    "            unk_token (str): the UNK token to add into the Vocabulary\r\n",
    "            \r\n",
    "        \"\"\"\r\n",
    "\r\n",
    "        if token_to_idx is None:\r\n",
    "            token_to_idx = {}\r\n",
    "        self._token_to_idx = token_to_idx\r\n",
    "\r\n",
    "        self._idx_to_token = {idx: token \r\n",
    "                              for token, idx in self._token_to_idx.items()}\r\n",
    "        \r\n",
    "        self._add_unk = add_unk\r\n",
    "        self._unk_token = unk_token\r\n",
    "        self._mask_token = mask_token\r\n",
    "        \r\n",
    "        self.mask_index = self.add_token(self._mask_token)\r\n",
    "        self.unk_index = -1\r\n",
    "        if add_unk:\r\n",
    "            self.unk_index = self.add_token(unk_token) \r\n",
    "        \r\n",
    "    def to_serializable(self):\r\n",
    "        \"\"\" returns a dictionary that can be serialized \"\"\"\r\n",
    "        return {'token_to_idx': self._token_to_idx, \r\n",
    "                'add_unk': self._add_unk, \r\n",
    "                'unk_token': self._unk_token, \r\n",
    "                'mask_token': self._mask_token}\r\n",
    "\r\n",
    "    @classmethod\r\n",
    "    def from_serializable(cls, contents):\r\n",
    "        \"\"\" instantiates the Vocabulary from a serialized dictionary \"\"\"\r\n",
    "        return cls(**contents)\r\n",
    "\r\n",
    "    def add_token(self, token):\r\n",
    "        \"\"\"Update mapping dicts based on the token.\r\n",
    "\r\n",
    "        Args:\r\n",
    "            token (str): the item to add into the Vocabulary\r\n",
    "        Returns:\r\n",
    "            index (int): the integer corresponding to the token\r\n",
    "        \"\"\"\r\n",
    "        if token in self._token_to_idx:\r\n",
    "            index = self._token_to_idx[token]\r\n",
    "        else:\r\n",
    "            index = len(self._token_to_idx)\r\n",
    "            self._token_to_idx[token] = index\r\n",
    "            self._idx_to_token[index] = token\r\n",
    "        return index\r\n",
    "            \r\n",
    "    def add_many(self, tokens):\r\n",
    "        \"\"\"Add a list of tokens into the Vocabulary\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            tokens (list): a list of string tokens\r\n",
    "        Returns:\r\n",
    "            indices (list): a list of indices corresponding to the tokens\r\n",
    "        \"\"\"\r\n",
    "        return [self.add_token(token) for token in tokens]\r\n",
    "\r\n",
    "    def lookup_token(self, token):\r\n",
    "        \"\"\"Retrieve the index associated with the token \r\n",
    "          or the UNK index if token isn't present.\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            token (str): the token to look up \r\n",
    "        Returns:\r\n",
    "            index (int): the index corresponding to the token\r\n",
    "        Notes:\r\n",
    "            `unk_index` needs to be >=0 (having been added into the Vocabulary) \r\n",
    "              for the UNK functionality \r\n",
    "        \"\"\"\r\n",
    "        if self.unk_index >= 0:\r\n",
    "            return self._token_to_idx.get(token, self.unk_index)\r\n",
    "        else:\r\n",
    "            return self._token_to_idx[token]\r\n",
    "\r\n",
    "    def lookup_index(self, index):\r\n",
    "        \"\"\"Return the token associated with the index\r\n",
    "        \r\n",
    "        Args: \r\n",
    "            index (int): the index to look up\r\n",
    "        Returns:\r\n",
    "            token (str): the token corresponding to the index\r\n",
    "        Raises:\r\n",
    "            KeyError: if the index is not in the Vocabulary\r\n",
    "        \"\"\"\r\n",
    "        if index not in self._idx_to_token:\r\n",
    "            raise KeyError(\"the index (%d) is not in the Vocabulary\" % index)\r\n",
    "        return self._idx_to_token[index]\r\n",
    "\r\n",
    "    def __str__(self):\r\n",
    "        return \"<Vocabulary(size=%d)>\" % len(self)\r\n",
    "\r\n",
    "    def __len__(self):\r\n",
    "        return len(self._token_to_idx)"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Vectorizer\r\n",
    "\r\n",
    "The Vectorizer class build a vocabulary from the data frame using its class method `from_dataframe()`, then the `vectorize()` methods look up the input `contex` tokens, returns a list of indices of these tokens. "
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "source": [
    "class CBOWVectorizer(object):\r\n",
    "    \"\"\" The Vectorizer which coordinates the Vocabularies and puts them to use\"\"\"    \r\n",
    "    def __init__(self, cbow_vocab):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            cbow_vocab (Vocabulary): maps words to integers\r\n",
    "        \"\"\"\r\n",
    "        self.cbow_vocab = cbow_vocab\r\n",
    "\r\n",
    "    def vectorize(self, context, vector_length=-1):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            context (str): the string of words separated by a space\r\n",
    "            vector_length (int): an argument for forcing the length of index vector\r\n",
    "        \"\"\"\r\n",
    "\r\n",
    "        indices = [self.cbow_vocab.lookup_token(token) for token in context.split(' ')]\r\n",
    "        if vector_length < 0:\r\n",
    "            vector_length = len(indices)\r\n",
    "\r\n",
    "        out_vector = np.zeros(vector_length, dtype=np.int64)\r\n",
    "        out_vector[:len(indices)] = indices\r\n",
    "        out_vector[len(indices):] = self.cbow_vocab.mask_index\r\n",
    "\r\n",
    "        return out_vector\r\n",
    "    \r\n",
    "    @classmethod\r\n",
    "    def from_dataframe(cls, cbow_df):\r\n",
    "        \"\"\"Instantiate the vectorizer from the dataset dataframe\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            cbow_df (pandas.DataFrame): the target dataset\r\n",
    "        Returns:\r\n",
    "            an instance of the CBOWVectorizer\r\n",
    "        \"\"\"\r\n",
    "        cbow_vocab = Vocabulary()\r\n",
    "        for index, row in cbow_df.iterrows():\r\n",
    "            for token in row.context.split(' '):\r\n",
    "                cbow_vocab.add_token(token)\r\n",
    "            cbow_vocab.add_token(row.target)\r\n",
    "            \r\n",
    "        return cls(cbow_vocab)\r\n",
    "\r\n",
    "    @classmethod\r\n",
    "    def from_serializable(cls, contents):\r\n",
    "        cbow_vocab = \\\r\n",
    "            Vocabulary.from_serializable(contents['cbow_vocab'])\r\n",
    "        return cls(cbow_vocab=cbow_vocab)\r\n",
    "\r\n",
    "    def to_serializable(self):\r\n",
    "        return {'cbow_vocab': self.cbow_vocab.to_serializable()}"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### The Dataset with DataLoader\r\n",
    "\r\n",
    "The `Dataset` class uses class method `load_dataset_and_make_vectorizer()` to read the input data into a data frame, then pass the data frame to the `Vectorizer` class to populate a Vocabulary object such that when the `vectorize()` method is called, it will perform the appropriate vectorization, in this case, return the vocabulary indices of the input context words.\r\n",
    "\r\n",
    ":::{note}\r\n",
    "The `generate_batches()` method uses PyTorch build in `DataLoader` class to create mini-batches. \r\n",
    ":::"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "source": [
    "class CBOWDataset(Dataset):\r\n",
    "    def __init__(self, cbow_df, vectorizer):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            cbow_df (pandas.DataFrame): the dataset\r\n",
    "            vectorizer (CBOWVectorizer): vectorizer instatiated from dataset\r\n",
    "        \"\"\"\r\n",
    "        self.cbow_df = cbow_df\r\n",
    "        self._vectorizer = vectorizer\r\n",
    "        \r\n",
    "        measure_len = lambda context: len(context.split(\" \"))\r\n",
    "        self._max_seq_length = max(map(measure_len, cbow_df.context))\r\n",
    "        \r\n",
    "        self.train_df = self.cbow_df[self.cbow_df.split=='train']\r\n",
    "        self.train_size = len(self.train_df)\r\n",
    "\r\n",
    "        self.val_df = self.cbow_df[self.cbow_df.split=='val']\r\n",
    "        self.validation_size = len(self.val_df)\r\n",
    "\r\n",
    "        self.test_df = self.cbow_df[self.cbow_df.split=='test']\r\n",
    "        self.test_size = len(self.test_df)\r\n",
    "\r\n",
    "        self._lookup_dict = {'train': (self.train_df, self.train_size),\r\n",
    "                             'val': (self.val_df, self.validation_size),\r\n",
    "                             'test': (self.test_df, self.test_size)}\r\n",
    "\r\n",
    "        self.set_split('train')\r\n",
    "\r\n",
    "    @classmethod\r\n",
    "    def load_dataset_and_make_vectorizer(cls, cbow_csv):\r\n",
    "        \"\"\"Load dataset and make a new vectorizer from scratch\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            cbow_csv (str): location of the dataset\r\n",
    "        Returns:\r\n",
    "            an instance of CBOWDataset\r\n",
    "        \"\"\"\r\n",
    "        cbow_df = pd.read_csv(cbow_csv)\r\n",
    "        train_cbow_df = cbow_df[cbow_df.split=='train']\r\n",
    "        return cls(cbow_df, CBOWVectorizer.from_dataframe(train_cbow_df))\r\n",
    "\r\n",
    "    @classmethod\r\n",
    "    def load_dataset_and_load_vectorizer(cls, cbow_csv, vectorizer_filepath):\r\n",
    "        \"\"\"Load dataset and the corresponding vectorizer. \r\n",
    "        Used in the case in the vectorizer has been cached for re-use\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            cbow_csv (str): location of the dataset\r\n",
    "            vectorizer_filepath (str): location of the saved vectorizer\r\n",
    "        Returns:\r\n",
    "            an instance of CBOWDataset\r\n",
    "        \"\"\"\r\n",
    "        cbow_df = pd.read_csv(cbow_csv)\r\n",
    "        vectorizer = cls.load_vectorizer_only(vectorizer_filepath)\r\n",
    "        return cls(cbow_df, vectorizer)\r\n",
    "\r\n",
    "    @staticmethod\r\n",
    "    def load_vectorizer_only(vectorizer_filepath):\r\n",
    "        \"\"\"a static method for loading the vectorizer from file\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            vectorizer_filepath (str): the location of the serialized vectorizer\r\n",
    "        Returns:\r\n",
    "            an instance of CBOWVectorizer\r\n",
    "        \"\"\"\r\n",
    "        with open(vectorizer_filepath) as fp:\r\n",
    "            return CBOWVectorizer.from_serializable(json.load(fp))\r\n",
    "\r\n",
    "    def save_vectorizer(self, vectorizer_filepath):\r\n",
    "        \"\"\"saves the vectorizer to disk using json\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            vectorizer_filepath (str): the location to save the vectorizer\r\n",
    "        \"\"\"\r\n",
    "        with open(vectorizer_filepath, \"w\") as fp:\r\n",
    "            json.dump(self._vectorizer.to_serializable(), fp)\r\n",
    "\r\n",
    "    def get_vectorizer(self):\r\n",
    "        \"\"\" returns the vectorizer \"\"\"\r\n",
    "        return self._vectorizer\r\n",
    "        \r\n",
    "    def set_split(self, split=\"train\"):\r\n",
    "        \"\"\" selects the splits in the dataset using a column in the dataframe \"\"\"\r\n",
    "        self._target_split = split\r\n",
    "        self._target_df, self._target_size = self._lookup_dict[split]\r\n",
    "\r\n",
    "    def __len__(self):\r\n",
    "        return self._target_size\r\n",
    "\r\n",
    "    def __getitem__(self, index):\r\n",
    "        \"\"\"the primary entry point method for PyTorch datasets\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            index (int): the index to the data point \r\n",
    "        Returns:\r\n",
    "            a dictionary holding the data point's features (x_data) and label (y_target)\r\n",
    "        \"\"\"\r\n",
    "        row = self._target_df.iloc[index]\r\n",
    "\r\n",
    "        context_vector = \\\r\n",
    "            self._vectorizer.vectorize(row.context, self._max_seq_length)\r\n",
    "        target_index = self._vectorizer.cbow_vocab.lookup_token(row.target)\r\n",
    "\r\n",
    "        return {'x_data': context_vector,\r\n",
    "                'y_target': target_index}\r\n",
    "\r\n",
    "    def get_num_batches(self, batch_size):\r\n",
    "        \"\"\"Given a batch size, return the number of batches in the dataset\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            batch_size (int)\r\n",
    "        Returns:\r\n",
    "            number of batches in the dataset\r\n",
    "        \"\"\"\r\n",
    "        return len(self) // batch_size\r\n",
    "    \r\n",
    "def generate_batches(dataset, batch_size, shuffle=True,\r\n",
    "                     drop_last=True, device=\"cpu\"): \r\n",
    "    \"\"\"\r\n",
    "    A generator function which wraps the PyTorch DataLoader. It will \r\n",
    "      ensure each tensor is on the write device location.\r\n",
    "    \"\"\"\r\n",
    "    dataloader = DataLoader(dataset=dataset, batch_size=batch_size,\r\n",
    "                            shuffle=shuffle, drop_last=drop_last)\r\n",
    "\r\n",
    "    for data_dict in dataloader:\r\n",
    "        out_data_dict = {}\r\n",
    "        for name, tensor in data_dict.items():\r\n",
    "            out_data_dict[name] = data_dict[name].to(device)\r\n",
    "        yield out_data_dict"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "## The Model: CBOW\r\n",
    "\r\n",
    "The CBOW model uses an embedding layer `nn.Embedding()` which will have weights to be intialised randomly and updated through training. These weights will be the embeddings for each token after training. The model is a simple one hidden layer fully connected network. "
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "source": [
    "class CBOWClassifier(nn.Module): # Simplified cbow Model\r\n",
    "    def __init__(self, vocabulary_size, embedding_size, padding_idx=0):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            vocabulary_size (int): number of vocabulary items, controls the\r\n",
    "                number of embeddings and prediction vector size\r\n",
    "            embedding_size (int): size of the embeddings\r\n",
    "            padding_idx (int): default 0; Embedding will not use this index\r\n",
    "        \"\"\"\r\n",
    "        super(CBOWClassifier, self).__init__()\r\n",
    "        \r\n",
    "        self.embedding =  nn.Embedding(num_embeddings=vocabulary_size, \r\n",
    "                                       embedding_dim=embedding_size,\r\n",
    "                                       padding_idx=padding_idx)\r\n",
    "        self.fc1 = nn.Linear(in_features=embedding_size,\r\n",
    "                             out_features=vocabulary_size)\r\n",
    "\r\n",
    "    def forward(self, x_in, apply_softmax=False):\r\n",
    "        \"\"\"The forward pass of the classifier\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            x_in (torch.Tensor): an input data tensor. \r\n",
    "                x_in.shape should be (batch, input_dim)\r\n",
    "            apply_softmax (bool): a flag for the softmax activation\r\n",
    "                should be false if used with the Cross Entropy losses\r\n",
    "        Returns:\r\n",
    "            the resulting tensor. tensor.shape should be (batch, output_dim)\r\n",
    "        \"\"\"\r\n",
    "        x_embedded_sum = F.dropout(self.embedding(x_in).sum(dim=1), 0.3)\r\n",
    "        y_out = self.fc1(x_embedded_sum)\r\n",
    "        \r\n",
    "        if apply_softmax:\r\n",
    "            y_out = F.softmax(y_out, dim=1)\r\n",
    "            \r\n",
    "        return y_out"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "## Training Routine"
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Helper functions"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "source": [
    "def make_train_state(args):\r\n",
    "    return {'stop_early': False,\r\n",
    "            'early_stopping_step': 0,\r\n",
    "            'early_stopping_best_val': 1e8,\r\n",
    "            'learning_rate': args.learning_rate,\r\n",
    "            'epoch_index': 0,\r\n",
    "            'train_loss': [],\r\n",
    "            'train_acc': [],\r\n",
    "            'val_loss': [],\r\n",
    "            'val_acc': [],\r\n",
    "            'test_loss': -1,\r\n",
    "            'test_acc': -1,\r\n",
    "            'model_filename': args.model_state_file}\r\n",
    "\r\n",
    "def update_train_state(args, model, train_state):\r\n",
    "    \"\"\"Handle the training state updates.\r\n",
    "\r\n",
    "    Components:\r\n",
    "     - Early Stopping: Prevent overfitting.\r\n",
    "     - Model Checkpoint: Model is saved if the model is better\r\n",
    "\r\n",
    "    :param args: main arguments\r\n",
    "    :param model: model to train\r\n",
    "    :param train_state: a dictionary representing the training state values\r\n",
    "    :returns:\r\n",
    "        a new train_state\r\n",
    "    \"\"\"\r\n",
    "\r\n",
    "    # Save one model at least\r\n",
    "    if train_state['epoch_index'] == 0:\r\n",
    "        torch.save(model.state_dict(), train_state['model_filename'])\r\n",
    "        train_state['stop_early'] = False\r\n",
    "\r\n",
    "    # Save model if performance improved\r\n",
    "    elif train_state['epoch_index'] >= 1:\r\n",
    "        loss_tm1, loss_t = train_state['val_loss'][-2:]\r\n",
    "\r\n",
    "        # If loss worsened\r\n",
    "        if loss_t >= train_state['early_stopping_best_val']:\r\n",
    "            # Update step\r\n",
    "            train_state['early_stopping_step'] += 1\r\n",
    "        # Loss decreased\r\n",
    "        else:\r\n",
    "            # Save the best model\r\n",
    "            if loss_t < train_state['early_stopping_best_val']:\r\n",
    "                torch.save(model.state_dict(), train_state['model_filename'])\r\n",
    "\r\n",
    "            # Reset early stopping step\r\n",
    "            train_state['early_stopping_step'] = 0\r\n",
    "\r\n",
    "        # Stop early ?\r\n",
    "        train_state['stop_early'] = \\\r\n",
    "            train_state['early_stopping_step'] >= args.early_stopping_criteria\r\n",
    "\r\n",
    "    return train_state\r\n",
    "\r\n",
    "def compute_accuracy(y_pred, y_target):\r\n",
    "    _, y_pred_indices = y_pred.max(dim=1)\r\n",
    "    n_correct = torch.eq(y_pred_indices, y_target).sum().item()\r\n",
    "    return n_correct / len(y_pred_indices) * 100"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "#### general utilities"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "source": [
    "def set_seed_everywhere(seed, cuda):\r\n",
    "    np.random.seed(seed)\r\n",
    "    torch.manual_seed(seed)\r\n",
    "    if cuda:\r\n",
    "        torch.cuda.manual_seed_all(seed)\r\n",
    "\r\n",
    "def handle_dirs(dirpath):\r\n",
    "    if not os.path.exists(dirpath):\r\n",
    "        os.makedirs(dirpath)"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Settings and some prep work"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "source": [
    "args = Namespace(\r\n",
    "    # Data and Path information\r\n",
    "    cbow_csv=\"../data/books/frankenstein_with_splits.csv\",\r\n",
    "    vectorizer_file=\"vectorizer.json\",\r\n",
    "    model_state_file=\"model.pth\",\r\n",
    "    save_dir=\"model_storage\\cbow\",\r\n",
    "    # Model hyper parameters\r\n",
    "    embedding_size=50,\r\n",
    "    # Training hyper parameters\r\n",
    "    seed=1337,\r\n",
    "    num_epochs=100,\r\n",
    "    learning_rate=0.0001,\r\n",
    "    batch_size=32,\r\n",
    "    early_stopping_criteria=5,\r\n",
    "    # Runtime options\r\n",
    "    cuda=True,\r\n",
    "    catch_keyboard_interrupt=True,\r\n",
    "    reload_from_files=False,\r\n",
    "    expand_filepaths_to_save_dir=True\r\n",
    ")\r\n",
    "\r\n",
    "if args.expand_filepaths_to_save_dir:\r\n",
    "    args.vectorizer_file = os.path.join(args.save_dir,\r\n",
    "                                        args.vectorizer_file)\r\n",
    "\r\n",
    "    args.model_state_file = os.path.join(args.save_dir,\r\n",
    "                                         args.model_state_file)\r\n",
    "    \r\n",
    "    print(\"Expanded filepaths: \")\r\n",
    "    print(\"\\t{}\".format(args.vectorizer_file))\r\n",
    "    print(\"\\t{}\".format(args.model_state_file))\r\n",
    "    \r\n",
    "\r\n",
    "# Check CUDA\r\n",
    "if not torch.cuda.is_available():\r\n",
    "    args.cuda = False\r\n",
    "\r\n",
    "args.device = torch.device(\"cuda\" if args.cuda else \"cpu\")\r\n",
    "    \r\n",
    "print(\"Using CUDA: {}\".format(args.cuda))\r\n",
    "\r\n",
    "\r\n",
    "# Set seed for reproducibility\r\n",
    "set_seed_everywhere(args.seed, args.cuda)\r\n",
    "\r\n",
    "# handle dirs\r\n",
    "handle_dirs(args.save_dir)"
   ],
   "outputs": [
    {
     "output_type": "stream",
     "name": "stdout",
     "text": [
      "Expanded filepaths: \n",
      "\tmodel_storage\\cbow\\vectorizer.json\n",
      "\tmodel_storage\\cbow\\model.pth\n",
      "Using CUDA: True\n"
     ]
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Initializations"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "source": [
    "if args.reload_from_files:\r\n",
    "    print(\"Loading dataset and loading vectorizer\")\r\n",
    "    dataset = CBOWDataset.load_dataset_and_load_vectorizer(args.cbow_csv,\r\n",
    "                                                           args.vectorizer_file)\r\n",
    "else:\r\n",
    "    print(\"Loading dataset and creating vectorizer\")\r\n",
    "    dataset = CBOWDataset.load_dataset_and_make_vectorizer(args.cbow_csv)\r\n",
    "    dataset.save_vectorizer(args.vectorizer_file)\r\n",
    "    \r\n",
    "vectorizer = dataset.get_vectorizer()\r\n",
    "\r\n",
    "classifier = CBOWClassifier(vocabulary_size=len(vectorizer.cbow_vocab), \r\n",
    "                            embedding_size=args.embedding_size)\r\n"
   ],
   "outputs": [
    {
     "output_type": "stream",
     "name": "stdout",
     "text": [
      "Loading dataset and creating vectorizer\n"
     ]
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Training loop\r\n",
    "\r\n",
    "This takes 904 seconds (15 mins). "
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "source": [
    "classifier = classifier.to(args.device)\r\n",
    "    \r\n",
    "loss_func = nn.CrossEntropyLoss()\r\n",
    "optimizer = optim.Adam(classifier.parameters(), lr=args.learning_rate)\r\n",
    "scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer=optimizer,\r\n",
    "                                                 mode='min', factor=0.5,\r\n",
    "                                                 patience=1)\r\n",
    "train_state = make_train_state(args)\r\n",
    "\r\n",
    "epoch_bar = tqdm(desc='training routine', \r\n",
    "                          total=args.num_epochs,\r\n",
    "                          position=0)\r\n",
    "\r\n",
    "dataset.set_split('train')\r\n",
    "train_bar = tqdm(desc='split=train',\r\n",
    "                          total=dataset.get_num_batches(args.batch_size), \r\n",
    "                          position=1, \r\n",
    "                          leave=True)\r\n",
    "dataset.set_split('val')\r\n",
    "val_bar = tqdm(desc='split=val',\r\n",
    "                        total=dataset.get_num_batches(args.batch_size), \r\n",
    "                        position=1, \r\n",
    "                        leave=True)\r\n",
    "\r\n",
    "try:\r\n",
    "    for epoch_index in range(args.num_epochs):\r\n",
    "        train_state['epoch_index'] = epoch_index\r\n",
    "\r\n",
    "        # Iterate over training dataset\r\n",
    "\r\n",
    "        # setup: batch generator, set loss and acc to 0, set train mode on\r\n",
    "\r\n",
    "        dataset.set_split('train')\r\n",
    "        batch_generator = generate_batches(dataset, \r\n",
    "                                           batch_size=args.batch_size, \r\n",
    "                                           device=args.device)\r\n",
    "        running_loss = 0.0\r\n",
    "        running_acc = 0.0\r\n",
    "        classifier.train()\r\n",
    "\r\n",
    "        for batch_index, batch_dict in enumerate(batch_generator):\r\n",
    "            # the training routine is these 5 steps:\r\n",
    "\r\n",
    "            # --------------------------------------\r\n",
    "            # step 1. zero the gradients\r\n",
    "            optimizer.zero_grad()\r\n",
    "\r\n",
    "            # step 2. compute the output\r\n",
    "            y_pred = classifier(x_in=batch_dict['x_data'])\r\n",
    "\r\n",
    "            # step 3. compute the loss\r\n",
    "            loss = loss_func(y_pred, batch_dict['y_target'])\r\n",
    "            loss_t = loss.item()\r\n",
    "            running_loss += (loss_t - running_loss) / (batch_index + 1)\r\n",
    "\r\n",
    "            # step 4. use loss to produce gradients\r\n",
    "            loss.backward()\r\n",
    "\r\n",
    "            # step 5. use optimizer to take gradient step\r\n",
    "            optimizer.step()\r\n",
    "            # -----------------------------------------\r\n",
    "            # compute the accuracy\r\n",
    "            acc_t = compute_accuracy(y_pred, batch_dict['y_target'])\r\n",
    "            running_acc += (acc_t - running_acc) / (batch_index + 1)\r\n",
    "\r\n",
    "            # update bar\r\n",
    "            train_bar.set_postfix(loss=running_loss, acc=running_acc, \r\n",
    "                            epoch=epoch_index)\r\n",
    "            train_bar.update()\r\n",
    "\r\n",
    "        train_state['train_loss'].append(running_loss)\r\n",
    "        train_state['train_acc'].append(running_acc)\r\n",
    "\r\n",
    "        # Iterate over val dataset\r\n",
    "\r\n",
    "        # setup: batch generator, set loss and acc to 0; set eval mode on\r\n",
    "        dataset.set_split('val')\r\n",
    "        batch_generator = generate_batches(dataset, \r\n",
    "                                           batch_size=args.batch_size, \r\n",
    "                                           device=args.device)\r\n",
    "        running_loss = 0.\r\n",
    "        running_acc = 0.\r\n",
    "        classifier.eval()\r\n",
    "\r\n",
    "        for batch_index, batch_dict in enumerate(batch_generator):\r\n",
    "\r\n",
    "            # compute the output\r\n",
    "            y_pred =  classifier(x_in=batch_dict['x_data'])\r\n",
    "\r\n",
    "            # step 3. compute the loss\r\n",
    "            loss = loss_func(y_pred, batch_dict['y_target'])\r\n",
    "            loss_t = loss.item()\r\n",
    "            running_loss += (loss_t - running_loss) / (batch_index + 1)\r\n",
    "\r\n",
    "            # compute the accuracy\r\n",
    "            acc_t = compute_accuracy(y_pred, batch_dict['y_target'])\r\n",
    "            running_acc += (acc_t - running_acc) / (batch_index + 1)\r\n",
    "            val_bar.set_postfix(loss=running_loss, acc=running_acc, \r\n",
    "                            epoch=epoch_index)\r\n",
    "            val_bar.update()\r\n",
    "\r\n",
    "        train_state['val_loss'].append(running_loss)\r\n",
    "        train_state['val_acc'].append(running_acc)\r\n",
    "\r\n",
    "        train_state = update_train_state(args=args, model=classifier,\r\n",
    "                                         train_state=train_state)\r\n",
    "\r\n",
    "        scheduler.step(train_state['val_loss'][-1])\r\n",
    "\r\n",
    "        if train_state['stop_early']:\r\n",
    "            break\r\n",
    "\r\n",
    "        train_bar.n = 0\r\n",
    "        val_bar.n = 0\r\n",
    "        epoch_bar.update()\r\n",
    "except KeyboardInterrupt:\r\n",
    "    print(\"Exiting loop\")\r\n"
   ],
   "outputs": [
    {
     "output_type": "display_data",
     "data": {
      "application/vnd.jupyter.widget-view+json": {
       "version_major": 2,
       "version_minor": 0,
       "model_id": "b209576af19948b2a87296e5cb15d7fe"
      },
      "text/plain": [
       "training routine:   0%|          | 0/100 [00:00<?, ?it/s]"
      ]
     },
     "metadata": {}
    },
    {
     "output_type": "display_data",
     "data": {
      "application/vnd.jupyter.widget-view+json": {
       "version_major": 2,
       "version_minor": 0,
       "model_id": "5524365a303b45a5a35b536f5d3264fc"
      },
      "text/plain": [
       "split=train:   0%|          | 0/1984 [00:00<?, ?it/s]"
      ]
     },
     "metadata": {}
    },
    {
     "output_type": "display_data",
     "data": {
      "application/vnd.jupyter.widget-view+json": {
       "version_major": 2,
       "version_minor": 0,
       "model_id": "f84efa8ee8a847cd8cca2c3556dca2ba"
      },
      "text/plain": [
       "split=val:   0%|          | 0/425 [00:00<?, ?it/s]"
      ]
     },
     "metadata": {}
    }
   ],
   "metadata": {
    "code_folding": []
   }
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "source": [
    "# compute the loss & accuracy on the test set using the best available model\r\n",
    "\r\n",
    "classifier.load_state_dict(torch.load(train_state['model_filename']))\r\n",
    "classifier = classifier.to(args.device)\r\n",
    "loss_func = nn.CrossEntropyLoss()\r\n",
    "\r\n",
    "dataset.set_split('test')\r\n",
    "batch_generator = generate_batches(dataset, \r\n",
    "                                   batch_size=args.batch_size, \r\n",
    "                                   device=args.device)\r\n",
    "running_loss = 0.\r\n",
    "running_acc = 0.\r\n",
    "classifier.eval()\r\n",
    "\r\n",
    "for batch_index, batch_dict in enumerate(batch_generator):\r\n",
    "    # compute the output\r\n",
    "    y_pred =  classifier(x_in=batch_dict['x_data'])\r\n",
    "    \r\n",
    "    # compute the loss\r\n",
    "    loss = loss_func(y_pred, batch_dict['y_target'])\r\n",
    "    loss_t = loss.item()\r\n",
    "    running_loss += (loss_t - running_loss) / (batch_index + 1)\r\n",
    "\r\n",
    "    # compute the accuracy\r\n",
    "    acc_t = compute_accuracy(y_pred, batch_dict['y_target'])\r\n",
    "    running_acc += (acc_t - running_acc) / (batch_index + 1)\r\n",
    "\r\n",
    "train_state['test_loss'] = running_loss\r\n",
    "train_state['test_acc'] = running_acc\r\n"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "source": [
    "print(\"Test loss: {};\".format(train_state['test_loss']))\r\n",
    "print(\"Test Accuracy: {}\".format(train_state['test_acc']))"
   ],
   "outputs": [
    {
     "output_type": "stream",
     "name": "stdout",
     "text": [
      "Test loss: 8.131561214222629;\n",
      "Test Accuracy: 11.551470588235299\n"
     ]
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    ":::{warning}\r\n",
    "The accuracy is pretty low, this is largely to do with the small training set - only one book. In reality, the model needs to be trained with at least terabytes of data. We have shown in the lectures on how model performace are significantly affected by the data size. The larger the dataset, the better the performance.\r\n",
    ":::"
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Trained Embeddings"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "source": [
    "def pretty_print(results):\r\n",
    "    \"\"\"\r\n",
    "    Pretty print embedding results.\r\n",
    "    \"\"\"\r\n",
    "    for item in results:\r\n",
    "        print (\"...[%.2f] - %s\"%(item[1], item[0]))\r\n",
    "\r\n",
    "def get_closest(target_word, word_to_idx, embeddings, n=5):\r\n",
    "    \"\"\"\r\n",
    "    Get the n closest\r\n",
    "    words to your word.\r\n",
    "    \"\"\"\r\n",
    "\r\n",
    "    # Calculate distances to all other words\r\n",
    "    \r\n",
    "    word_embedding = embeddings[word_to_idx[target_word.lower()]]\r\n",
    "    distances = []\r\n",
    "    for word, index in word_to_idx.items():\r\n",
    "        if word == \"<MASK>\" or word == target_word:\r\n",
    "            continue\r\n",
    "        distances.append((word, torch.dist(word_embedding, embeddings[index])))\r\n",
    "    \r\n",
    "    results = sorted(distances, key=lambda x: x[1])[1:n+2]\r\n",
    "    return results\r\n"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "source": [
    "word = 'monster'\r\n",
    "embeddings = classifier.embedding.weight.data\r\n",
    "word_to_idx = vectorizer.cbow_vocab._token_to_idx\r\n",
    "pretty_print(get_closest(word, word_to_idx, embeddings, n=5))"
   ],
   "outputs": [
    {
     "output_type": "stream",
     "name": "stdout",
     "text": [
      "...[7.32] - cares\n",
      "...[7.58] - griefs\n",
      "...[7.63] - sickness\n",
      "...[7.66] - trifling\n",
      "...[7.69] - saw\n",
      "...[7.70] - prolong\n"
     ]
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "source": [
    "target_words = ['frankenstein', 'monster', 'science', 'sickness', 'lonely', 'happy']\r\n",
    "\r\n",
    "embeddings = classifier.embedding.weight.data\r\n",
    "word_to_idx = vectorizer.cbow_vocab._token_to_idx\r\n",
    "\r\n",
    "for target_word in target_words: \r\n",
    "    print(f\"======={target_word}=======\")\r\n",
    "    if target_word not in word_to_idx:\r\n",
    "        print(\"Not in vocabulary\")\r\n",
    "        continue\r\n",
    "    pretty_print(get_closest(target_word, word_to_idx, embeddings, n=5))"
   ],
   "outputs": [
    {
     "output_type": "stream",
     "name": "stdout",
     "text": [
      "=======frankenstein=======\n",
      "...[6.93] - discrimination\n",
      "...[6.99] - slight\n",
      "...[7.02] - oppressive\n",
      "...[7.05] - spurned\n",
      "...[7.11] - illustrate\n",
      "...[7.11] - wandering\n",
      "=======monster=======\n",
      "...[7.32] - cares\n",
      "...[7.58] - griefs\n",
      "...[7.63] - sickness\n",
      "...[7.66] - trifling\n",
      "...[7.69] - saw\n",
      "...[7.70] - prolong\n",
      "=======science=======\n",
      "...[6.85] - mutual\n",
      "...[6.93] - mist\n",
      "...[6.95] - swelling\n",
      "...[7.01] - impression\n",
      "...[7.06] - darkened\n",
      "...[7.06] - nearly\n",
      "=======sickness=======\n",
      "...[6.37] - while\n",
      "...[6.45] - foundations\n",
      "...[6.61] - awoke\n",
      "...[6.65] - consoles\n",
      "...[6.70] - literally\n",
      "...[6.74] - depend\n",
      "=======lonely=======\n",
      "...[6.74] - unveiled\n",
      "...[6.88] - moonlight\n",
      "...[7.05] - ought\n",
      "...[7.08] - bed\n",
      "...[7.14] - superhuman\n",
      "...[7.14] - therefore\n",
      "=======happy=======\n",
      "...[6.25] - bottom\n",
      "...[6.42] - injury\n",
      "...[6.49] - chivalry\n",
      "...[6.50] - altered\n",
      "...[6.51] - penetrated\n",
      "...[6.54] - danger\n"
     ]
    }
   ],
   "metadata": {}
  }
 ],
 "metadata": {
  "_draft": {
   "nbviewer_url": "https://gist.github.com/7f86253a46c5444491fbb7541857edfa"
  },
  "gist": {
   "data": {
    "description": "chapters/06_RNN/tweet_classifcation/tweeter_classification.ipynb",
    "public": true
   },
   "id": "7f86253a46c5444491fbb7541857edfa"
  },
  "hide_input": false,
  "kernelspec": {
   "name": "python3",
   "display_name": "Python 3.8.10 64-bit ('cits4012': conda)"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.10"
  },
  "latex_envs": {
   "LaTeX_envs_menu_present": true,
   "bibliofile": "biblio.bib",
   "cite_by": "apalike",
   "current_citInitial": 1,
   "eqLabelWithNumbers": true,
   "eqNumInitial": 1,
   "hotkeys": {
    "equation": "Ctrl-E",
    "itemize": "Ctrl-I"
   },
   "labels_anchors": false,
   "latex_user_defs": false,
   "report_style_numbering": false,
   "user_envs_cfg": false
  },
  "nbTranslate": {
   "displayLangs": [
    "*"
   ],
   "hotkey": "alt-t",
   "langInMainMenu": true,
   "sourceLang": "en",
   "targetLang": "fr",
   "useGoogleTranslate": true
  },
  "notify_time": "30",
  "toc": {
   "colors": {
    "hover_highlight": "#DAA520",
    "running_highlight": "#FF0000",
    "selected_highlight": "#FFD700"
   },
   "moveMenuLeft": true,
   "nav_menu": {
    "height": "156px",
    "width": "252px"
   },
   "navigate_menu": true,
   "number_sections": true,
   "sideBar": false,
   "threshold": 4,
   "toc_cell": true,
   "toc_position": {
    "height": "255px",
    "left": "561px",
    "right": "20px",
    "top": "179px",
    "width": "266px"
   },
   "toc_section_display": "block",
   "toc_window_display": false
  },
  "interpreter": {
   "hash": "d990147e05fc0cc60dd3871899a6233eb6a5324c1885ded43d013dc915f7e535"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}